'''Utilities for processing a continuous stream of data in a synchronous context.

Please refer to the async counterpart in the module `mpservice.async_streamer`.
'''

import functools
import inspect
import logging
import multiprocessing
import random
from concurrent.futures import Future
from queue import Queue, Empty
from threading import Lock, Thread
from time import sleep
from typing import (
    Callable, TypeVar, Union, Optional,
    Iterable, Iterator,
    Tuple, List,
)


MAX_THREADS = min(32, multiprocessing.cpu_count() + 4)

NO_MORE_DATA = object()

logger = logging.getLogger(__name__)

T = TypeVar('T')
TT = TypeVar('TT')


def _is_exc(e):
    return isinstance(e, Exception) or (
        inspect.isclass(e) and issubclass(e, Exception)
    )


class IterQueue(Queue):
    DEFAULT_MAXSIZE = 256

    def __init__(self, maxsize: int = None, q_err: Queue = None):
        super().__init__(maxsize or self.DEFAULT_MAXSIZE)
        self._q_err = Queue() if q_err is None else q_err
        self._closed = False
        self._exhausted = False

    def put_end(self):
        assert not self._closed
        self.put(NO_MORE_DATA)
        self._closed = True

    def put_error(self, e):
        assert not self._closed
        self._q_err.put(e)

    def put(self, x, **kwargs):
        assert not self._closed
        super().put(x, **kwargs)

    def put_nowait(self, x):
        assert not self._closed
        super().put_nowait(x)

    def get(self, **kwargs):
        if self._exhausted:
            return NO_MORE_DATA
        z = super().get(**kwargs)
        if z is NO_MORE_DATA:
            self._exhausted = True
        return z

    def get_nowait(self):
        if not self._q_err.empty():
            raise self._q_err.get()
        if self._exhausted:
            return NO_MORE_DATA
        z = super().get_nowait()
        if z is NO_MORE_DATA:
            self._exhausted = True
        return z

    def __next__(self):
        while True:
            try:
                z = self.get_nowait()
                if z is NO_MORE_DATA:
                    raise StopIteration
                return z
            except Empty:
                sleep(0.005)

    def __iter__(self):
        return self


def stream(x: Iterable, maxsize: int = None) -> IterQueue:
    if isinstance(x, IterQueue):
        return x

    if not hasattr(x, '__iter__'):
        if hasattr(x, '__next__'):
            def f(x):
                while True:
                    try:
                        yield x.__next__()
                    except StopIteration:
                        break
            x = f(x)
        else:
            raise TypeError('`x` is not iterable')

    def enqueue(q_in, q_out):
        try:
            for v in q_in:
                q_out.put(v)
        except Exception as e:
            q_out.put_error(e)
        q_out.put_end()

    q = IterQueue(maxsize)
    t = Thread(target=enqueue, args=(x, q))
    t.start()

    return q


def collect(in_stream: Iterable[T]) -> List[T]:
    return list(in_stream)


def batch(q_in: IterQueue, q_out: IterQueue, batch_size: int) -> None:
    '''Take elements from an input stream,
    and bundle them up into batches up to a size limit,
    and produce the batches in an iterable.

    The output batches are all of the specified size,
    except possibly the final batch.
    There is no 'timeout' logic to produce a smaller batch.
    For efficiency, this requires the input stream to have a steady supply.
    If that is a concern, having a `buffer` on the input stream may help.
    '''
    assert 0 < batch_size <= 10000
    batch_ = []
    n = 0
    try:
        for x in q_in:
            batch_.append(x)
            n += 1
            if n >= batch_size:
                q_out.put(batch_)
                batch_ = []
                n = 0
        if n:
            q_out.put(batch_)
    except Exception as e:
        q_out.put_error(e)
    q_out.put_end()


def unbatch(q_in: IterQueue, q_out: IterQueue) -> None:
    '''Reverse of "batch", turning a stream of batches into
    a stream of individual elements.
    '''
    try:
        for batch in q_in:
            for x in batch:
                q_out.put(x)
    except Exception as e:
        q_out.put_error(e)
    q_out.put_end()


def buffer(q_in: IterQueue, q_out: IterQueue) -> None:
    '''Buffer is used to stabilize and improve the speed of data flow.

    A buffer is useful after any operation that can not guarantee
    (almost) instant availability of output. A buffer allows its
    output to "pile up" when the downstream consumer is slow in requests,
    so that data *is* available when the downstream does come to request
    data. The buffer evens out unstabilities in the speeds of upstream
    production and downstream consumption.
    '''
    for x in q_in:
        q_out.put(x)
    q_out.put_end()


def drop_if(q_in: IterQueue, q_out: IterQueue,
            func: Callable[[int, T], bool]) -> None:
    n = 0
    try:
        for x in q_in:
            if func(n, x):
                n += 1
                continue
            q_out.put(x)
            n += 1
    except Exception as e:
        q_out.put_error(e)
    q_out.put_end()


def drop_exceptions(q_in, q_out):
    return drop_if(q_in, q_out, lambda i, x: _is_exc(x))


def drop_first_n(q_in, q_out, n: int):
    assert n >= 0
    return drop_if(q_in, q_out, lambda i, x: i < n)


def keep_if(q_in: IterQueue,
            q_out: IterQueue,
            func: Callable[[int, T], bool]) -> None:
    n = 0
    try:
        for x in q_in:
            if func(n, x):
                q_out.put(x)
            n += 1
    except Exception as e:
        q_out.put_error(e)
    q_out.put_end()


def keep_every_nth(q_in, q_out, nth: int):
    assert nth > 0
    return keep_if(q_in, q_out, lambda i, x: i % nth == 0)


def keep_random(q_in, q_out, frac: float):
    assert 0 < frac <= 1
    rand = random.random
    return keep_if(q_in, q_out, lambda i, x: rand() < frac)


def keep_first_n(q_in, q_out, n: int):
    assert n > 0
    k = 0
    for x in q_in:
        k += 1
        if k > n:
            break
        q_out.put(x)
    q_out.put_end()


def peek_if(q_in: IterQueue,
            q_out: IterQueue,
            condition_func: Callable[[int, T], bool],
            peek_func: Callable[[int, T], None] = None,
            ) -> None:
    '''Take a peek at the data elements that statisfy the specified condition.

    `peek_func` usually prints out info of the data element,
    but can save it to a file or does other things. This happens *before*
    the element is sent downstream.

    The peek function usually should not modify the data element.
    '''
    if peek_func is None:
        def peek_func(i, x):
            print('')
            print('#', i)
            print(x)

    n = 0
    try:
        for x in q_in:
            if condition_func(n, x):
                peek_func(n, x)
            q_out.put(x)
            n += 1
    except Exception as e:
        q_out.put_error(e)
    q_out.put_end()


def peek_every_nth(q_in, q_out, nth: int, peek_func=None):
    return peek_if(q_in, q_out, lambda i, x: i % nth == 0, peek_func)


def peek_random(q_in, q_out, frac: float, peek_func=None):
    assert 0 < frac <= 1
    rand = random.random
    return peek_if(q_in, q_out, lambda i, x: rand() < frac, peek_func)


def log_every_nth(q_in, q_out, nth: int):
    def peek_func(i, x):
        logger.info('data item #%d:  %s', i, x)

    return peek_every_nth(q_in, q_out, nth, peek_func)


def log_exceptions(q_in, q_out, level: str = 'error'):
    flog = getattr(logger, level)
    return peek_if(
        q_in, q_out,
        lambda i, x: _is_exc(x),
        lambda i, x: flog(x)
    )


def transform(q_in: IterQueue,
              q_out: IterQueue,
              func: Callable[[T], TT],
              *,
              workers: Optional[Union[int, str]] = None,
              return_exceptions: bool = False,
              **func_args,
              ) -> None:
    '''Apply a transformation on each element of the data stream,
    producing a stream of corresponding results.

    `func`: a function that takes a single input item
    as the first positional argument and produces a result.
    Additional keywargs can be passed in via the keyward arguments
    `func_args`.

    The outputs are in the order of the input elements.

    `workers`: max number of concurrent calls to `func`. By default
    this is 1, i.e. there is no concurrency.
    '''
    if workers is None:
        workers = 1
    elif isinstance(workers, str):
        assert workers == 'max'
        workers = MAX_THREADS
    else:
        workers > 0

    if workers == 1:
        for x in q_in:
            try:
                z = func(x, **func_args)
                q_out.put(z)
            except Exception as e:
                if return_exceptions:
                    q_out.put(e)
                else:
                    q_out.put_error(e)
                    break  # No need to process subsequent data.
        q_out.put_end()
        return

    finished = False

    def _process(in_stream, out_stream, func, lock, **kwargs):
        nonlocal finished
        while not finished:
            with lock:
                if finished:
                    return
                try:
                    x = in_stream.__next__()
                    fut = Future()
                    out_stream.put(fut)
                except StopIteration:
                    finished = True
                    out_stream.put(NO_MORE_DATA)
                    return
                except Exception as e:
                    fut = Future()
                    out_stream.put(fut)
                    fut.set_exception(e)
                    continue

            try:
                y = func(x, **kwargs)
                fut.set_result(y)
            except Exception as e:
                fut.set_exception(e)

    out_stream = IterQueue(max(q_in.maxsize, workers * 8))
    lock = Lock()

    t_workers = [
        Thread(target=_process,
               args=(
                   q_in,
                   out_stream,
                   func,
                   lock,
               ),
               kwargs=func_args)
        for _ in range(workers)
    ]
    for t in t_workers:
        t.start()

    for fut in out_stream:
        try:
            z = fut.result()
            q_out.put(z)
        except Exception as e:
            if return_exceptions:
                q_out.put(e)
            else:
                q_out.put_error(e)
                finished = True
                break  # No need to process subsequent data.
    q_out.put_end()
    for t in t_workers:
        t.join()


def drain(q_in: IterQueue, log_nth: int = 0) -> Union[int, Tuple[int, int]]:
    '''Drain off the stream.

    Return the number of elements processed.
    '''
    if log_nth > 0:
        q_out = IterQueue(q_in.maxsize, q_in._q_err)
        t = Thread(target=log_every_nth, args=(q_in, q_out, log_nth))
        t.start()
    else:
        q_out = q_in
    n = 0
    nexc = 0
    for v in q_out:
        n += 1
        if _is_exc(v):
            nexc += 1
    if log_nth > 0:
        t.join()
    if nexc:
        return n, nexc
    return n


class Stream:
    @classmethod
    def registerapi(cls,
                    func: Callable[..., None],
                    *,
                    name: str = None,
                    maxsize: bool = False,
                    ) -> None:
        '''
        `func` expects the input and output streams (both of type IterQueue)
        as the first two positional arguments. It may take additional
        positional and keyword arguments. See the functions `batch`, `drop_if`,
        `transform`, etc for examples.

        The created method accepts the extra positional and keyword
        args after the first two positional args. The input and output
        streams are not args of the method, because the first is
        provided by the host object, whereas the second is constructed
        during this registration.

        If `maxsize` is `True`, the created method also takes keyword
        arg `maxsize`, which is passed to the constructor of the
        output stream object, which is of type IterQueue.

        User can use this method to register other functions so that they
        can be used as methods of a `Stream` object, just like `batch`,
        `drop_if`, etc.
        '''
        if not name:
            name = func.__name__

        if maxsize:
            @functools.wraps(func)
            def wrapped(self, *args, maxsize: int = None, **kwargs):
                if maxsize is None:
                    maxsize = self.in_stream.maxsize
                q_out = IterQueue(maxsize, self.in_stream._q_err)
                t = Thread(target=func,
                           args=(self.in_stream, q_out, *args),
                           kwargs=kwargs)
                t.start()
                return cls(q_out)
        else:
            @functools.wraps(func)
            def wrapped(self, *args, **kwargs):
                q_out = IterQueue(self.in_stream.maxsize,
                                  self.in_stream._q_err)
                t = Thread(target=func,
                           args=(self.in_stream, q_out, *args),
                           kwargs=kwargs)
                t.start()
                return cls(q_out)

        setattr(cls, name, wrapped)

    def __init__(self,
                 in_stream: Union[Iterable, Iterator, IterQueue],
                 maxsize: int = None):
        self.in_stream = stream(in_stream, maxsize)

    def __next__(self):
        return self.in_stream.__next__()

    def __iter__(self):
        return self.in_stream.__iter__()

    def collect(self):
        return collect(self.in_stream)

    def drain(self, log_nth: int = 0):
        return drain(self.in_stream, log_nth)

    def buffer(self, buffer_size: int = None):
        if buffer_size is None:
            buffer_size = self.in_stream.maxsize * 4
        q = IterQueue(buffer_size, self.in_stream._q_err)
        t = Thread(target=buffer, args=(self.in_stream, q))
        t.start()
        return self.__class__(q)


Stream.registerapi(batch, maxsize=True)
Stream.registerapi(unbatch, maxsize=True)
Stream.registerapi(drop_if)
Stream.registerapi(drop_exceptions)
Stream.registerapi(drop_first_n)
Stream.registerapi(keep_if)
Stream.registerapi(keep_every_nth)
Stream.registerapi(keep_random)
Stream.registerapi(keep_first_n)
Stream.registerapi(peek_if)
Stream.registerapi(peek_every_nth)
Stream.registerapi(peek_random)
Stream.registerapi(log_every_nth)
Stream.registerapi(log_exceptions)
Stream.registerapi(transform)



'''Utilities for processing a continuous stream of data in an async context.

There is an input stream, which is an AsyncIterator.
This stream goes through a series of async operations, each of which
takes an AsyncIterator, and returns another AsyncIterator.
Thanks to this consistency of input/output, the operations can be "chained".

If the input is not an AsyncIterator, but rather some other (sync or async)
iterable, then the function `stream` will turn it into an AsyncIterator.

The target use case is that one or more operations is I/O bound,
hence can benefit from async or multi-thread concurrency.
Concurrency is enabled by `transform`, `unordered_transform`, and `drain`.

There are two ways to use these utilities. In the first way, one calls the
operation functions in sequence, e.g.

    data = range(100)
    data_stream = stream(data)
    result = transform(
        batch(data_stream, 10),
        my_op_that_takes_stream_of_batches)
    result = [_ async for _ in unbatch(result))]

In the second way, one starts with a `Stream` object, and calls
its methods in a "chained" fashion:

    data = range(100)
    result = (
        Stream(data)
        .batch(10)
        .transform(my_op_that_takes_stream_of_batches)
        .unbatch()
        .collect()
        )
    result = await result

(Of course, you don't have to call all the methods in one statement.)

In most cases, the second way is recommended.

Reference for an early version: https://zpz.github.io/blog/stream-processing/
'''

import asyncio
import functools
import inspect
import logging
import multiprocessing
import random
from typing import (
    Callable, Awaitable, TypeVar, Optional, Union,
    AsyncIterable, AsyncIterator, Iterable,
    Tuple, List)


MAX_THREADS = min(32, multiprocessing.cpu_count() + 4)
# This default is suitable for I/O bound operations.
# For others, user may want to specify a smaller value.


logger = logging.getLogger(__name__)


NO_MORE_DATA = object()

T = TypeVar('T')
TT = TypeVar('TT')

# Iterable vs iterator
#
# if we need to use
#
#   for v in X:
#       ...
#
# then `X.__iter__()` is called to get an "iterator".
# In this case, `X` is an "iterable", and it must implement `__iter__`.
#
# If we do not use it that way, but rather only directly call
#
#   next(X)
#
# then `X` must implement `__next__` (but does not need to implement `__iter__`).
# This `X` is an "iterator".
#
# Often we let `__iter__` return `self`, and implement `__next__` in the same class.
# This way the class is both an iterable and an iterator.

# Async generator returns an async iterator.


def _is_exc(e):
    return isinstance(e, Exception) or (
        inspect.isclass(e) and issubclass(e, Exception))


def stream(x: Union[Iterable[T], AsyncIterable[T]]) -> AsyncIterator[T]:
    '''Turn a sync iterable into an async iterator.
    However, user should try to provide a natively async iterable
    if possible.

    The returned object has both `__anext__` and `__aiter__`
    methods.
    '''
    async def f1(data):
        async for v in data:
            yield v

    async def f2(data):
        while True:
            try:
                yield await data.__anext__()
            except StopAsyncIteration:
                break

    async def f3(data):
        for v in data:
            yield v

    async def f4(data):
        while True:
            try:
                yield data.__next__()
            except StopIteration:
                break

    if hasattr(x, '__aiter__'):
        if hasattr(x, '__anext__'):
            return x
        else:
            return f1(x)
    if hasattr(x, '__anext__'):
        return f2(x)
    if hasattr(x, '__iter__'):
        return f3(x)
    if hasattr(x, '__next__'):
        return f4(x)

    raise TypeError("`x` is neither iterable or async iterable")


async def collect(in_stream: AsyncIterable[T]) -> List[T]:
    return [_ async for _ in in_stream]


async def batch(in_stream: AsyncIterable[T],
                batch_size: int) -> AsyncIterator[List[T]]:
    '''Take elements from an input stream,
    and bundle them up into batches up to a size limit,
    and produce the batches in an iterable.

    The output batches are all of the specified size, except possibly the final batch.
    There is no 'timeout' logic to produce a smaller batch.
    For efficiency, this requires the input stream to have a steady supply.
    If that is a concern, having a `buffer` on the input stream may help.
    '''
    assert 0 < batch_size <= 10000
    batch_ = []
    n = 0
    async for x in in_stream:
        batch_.append(x)
        n += 1
        if n >= batch_size:
            yield batch_
            batch_ = []
            n = 0
    if n:
        yield batch_


async def unbatch(in_stream: AsyncIterable[Iterable[T]]) -> AsyncIterator[T]:
    '''Reverse of "batch", turning a stream of batches into
    a stream of individual elements.
    '''
    async for batch in in_stream:
        for x in batch:
            yield x


async def buffer(in_stream: AsyncIterable[T],
                 buffer_size: int = None) -> AsyncIterator[T]:
    '''Buffer is used to stabilize and improve the speed of data flow.

    A buffer is useful after any operation that can not guarantee
    (almost) instant availability of output. A buffer allows its
    output to "pile up" when the downstream consumer is slow in requests,
    so that data *is* available when the downstream does come to request
    data. The buffer evens out unstabilities in the speeds of upstream
    production and downstream consumption.
    '''
    out_stream = asyncio.Queue(maxsize=buffer_size or 256)

    async def buff(in_stream, out_stream):
        async for x in in_stream:
            await out_stream.put(x)
        await out_stream.put(NO_MORE_DATA)

    t = asyncio.create_task(buff(in_stream, out_stream))

    while True:
        if t.done() and t.exception() is not None:
            raise t.exception()
        x = await out_stream.get()
        if x is NO_MORE_DATA:
            break
        yield x

    await t


async def drop_if(in_stream: AsyncIterable[T],
                  func: Callable[[int, T], bool]) -> AsyncIterator[T]:
    n = 0
    async for x in in_stream:
        if func(n, x):
            n += 1
            continue
        yield x
        n += 1


def drop_exceptions(in_stream):
    return drop_if(in_stream, lambda i, x: _is_exc(x))


def drop_first_n(in_stream, n: int):
    assert n >= 0
    if n == 0:
        return in_stream
    return drop_if(in_stream, lambda i, x: i < n)


async def keep_if(in_stream: AsyncIterable[T],
                  func: Callable[[int, T], bool]) -> AsyncIterator[T]:
    n = 0
    async for x in in_stream:
        if func(n, x):
            yield x
        n += 1


def keep_every_nth(in_stream, nth: int):
    assert nth > 0
    return keep_if(in_stream, lambda i, x: i % nth == 0)


def keep_random(in_stream, frac: float):
    assert 0 < frac <= 1
    rand = random.random
    return keep_if(in_stream, lambda i, x: rand() < frac)


async def keep_first_n(in_stream, n: int):
    assert n > 0
    k = 0
    async for x in in_stream:
        yield x
        k += 1
        if k >= n:
            break


def _default_peek_func(i, x):
    print('')
    print('#', i)
    print(x)


async def peek_if(in_stream: AsyncIterable[T],
                  condition_func: Callable[[int, T], bool],
                  peek_func: Callable[[int, T], None] = None,
                  ) -> AsyncIterator[T]:
    '''Take a peek at the data elements that statisfy the specified condition.

    `peek_func` usually prints out info of the data element,
    but can save it to a file or does other things. This happens *before*
    the element is sent downstream.

    The peek function usually should not modify the data element.
    '''
    if peek_func is None:
        peek_func = _default_peek_func

    n = 0
    async for x in in_stream:
        if condition_func(n, x):
            peek_func(n, x)
        yield x
        n += 1


def peek_every_nth(in_stream, nth: int, peek_func=None):
    return peek_if(in_stream, lambda i, x: i % nth == 0, peek_func)


def peek_random(in_stream, frac: float, peek_func=None):
    assert 0 < frac <= 1
    rand = random.random
    return peek_if(in_stream, lambda i, x: rand() < frac, peek_func)


def log_every_nth(in_stream, nth: int):
    def peek_func(i, x):
        logger.info('data item #%d:  %s', i, x)

    return peek_every_nth(in_stream, nth, peek_func)


def log_exceptions(in_stream, level: str = 'error'):
    flog = getattr(logger, level)
    return peek_if(in_stream,
                   lambda i, x: _is_exc(x),
                   lambda i, x: flog(x))


# TODO: support sync function.
async def transform(
    in_stream: AsyncIterator[T],
    func: Callable[[T], Awaitable[TT]],
    *,
    workers: Optional[Union[int, str]] = None,
    return_exceptions: bool = False,
    **func_args,
) -> AsyncIterator[TT]:
    '''Apply a transformation on each element of the data stream,
    producing a stream of corresponding results.

    `func`: an async function that takes a single input item
    as the first positional argument and produces a result.
    Additional keywargs can be passed in via the keyward arguments
    `func_args`.

    The outputs are in the order of the input elements in `in_stream`.

    `workers`: max number of concurrent calls to `func`. By default
    this is 1, i.e. there is no concurrency.
    '''
    if workers is None:
        workers = 1
    elif isinstance(workers, str):
        assert workers == 'max'
        workers = MAX_THREADS
    else:
        workers > 0

    if workers == 1:
        async for x in in_stream:
            try:
                z = await func(x, **func_args)
                yield z
            except Exception as e:
                if return_exceptions:
                    yield e
                else:
                    raise e
        return

    finished = False

    async def _process(in_stream, lock, out_stream, func, **kwargs):
        nonlocal finished
        while not finished:
            async with lock:
                if finished:
                    return
                try:
                    x = await in_stream.__anext__()
                    fut = asyncio.Future()
                    await out_stream.put(fut)
                except StopAsyncIteration:
                    finished = True
                    await out_stream.put(NO_MORE_DATA)
                    return
                except Exception as e:
                    fut = asyncio.Future()
                    await out_stream.put(fut)
                    fut.set_exception(e)
                    continue

            try:
                y = await func(x, **kwargs)
                fut.set_result(y)
            except Exception as e:
                fut.set_exception(e)

    out_buffer_size = workers * 8
    out_stream = asyncio.Queue(out_buffer_size)
    lock = asyncio.Lock()

    t_workers = [
        asyncio.create_task(_process(
            in_stream,
            lock,
            out_stream,
            func,
            **func_args,
        ))
        for _ in range(workers)
    ]

    while True:
        fut = await out_stream.get()
        if fut is NO_MORE_DATA:
            break
        try:
            z = await fut
            yield z
        except Exception as e:
            if return_exceptions:
                yield e
            else:
                raise e

    for t in t_workers:
        await t


async def drain(in_stream: AsyncIterable,
                log_nth: int = 0) -> Union[int, Tuple[int, int]]:
    '''Drain off the stream.

    Return the number of elements processed.
    '''
    if log_nth:
        in_stream = log_every_nth(in_stream, log_nth)
    n = 0
    nexc = 0
    async for v in in_stream:
        n += 1
        if _is_exc(v):
            nexc += 1
    if nexc:
        return n, nexc
    return n


class Stream:
    @classmethod
    def registerapi(cls,
                    func: Callable[..., AsyncIterator],
                    *,
                    name: str = None) -> None:
        '''
        `func` expects the data stream, an AsyncIterable or AsyncIterator,
        as the first positional argument. It may take additional positional
        and keyword arguments. See the functions `batch`, `drop_if`,
        `transform`, etc for examples.

        User can use this method to register other functions so that they
        can be used as methods of a `Stream` object, just like `batch`,
        `drop_if`, etc.
        '''
        if not name:
            name = func.__name__

        @functools.wraps(func)
        def wrapped(self, *args, **kwargs):
            return cls(func(self.in_stream, *args, **kwargs))

        setattr(cls, name, wrapped)

    def __init__(self, in_stream: Union[Iterable, AsyncIterable]):
        self.in_stream = stream(in_stream)

    def __anext__(self):
        return self.in_stream.__anext__()

    def __aiter__(self):
        return self.in_stream.__aiter__()

    def collect(self):
        return collect(self.in_stream)

    def drain(self, log_nth=0):
        return drain(self.in_stream, log_nth)


Stream.registerapi(batch)
Stream.registerapi(unbatch)
Stream.registerapi(buffer)
Stream.registerapi(drop_if)
Stream.registerapi(drop_exceptions)
Stream.registerapi(drop_first_n)
Stream.registerapi(keep_if)
Stream.registerapi(keep_every_nth)
Stream.registerapi(keep_random)
Stream.registerapi(keep_first_n)
Stream.registerapi(peek_if)
Stream.registerapi(peek_every_nth)
Stream.registerapi(peek_random)
Stream.registerapi(log_every_nth)
Stream.registerapi(log_exceptions)
Stream.registerapi(transform)



from multiprocessing.managers import BaseManager


class ServerProcess:
    # User should subclass this class to implement the
    # functionalities they need.
    # An instance of this class is created in a "server process"
    # and serves as shared data for other processes.
    #
    # Usage:
    #
    #   (1) in main process,
    #
    #       obj = ServerProcess.start(...)
    #
    #   (2) pass `obj` to other processes
    #
    #   (3) in other processes, cal public methods of `obj`.
    #
    # `obj` is NOT an instance of the class `ServerProcess`.
    # It's a "proxy" object, which is like a reference to a
    # `ServerProcess` object in the "server process".
    # All public methods of `ServerProcess` can be used on this
    # proxy object from other processes.
    # Input and output should all be small, pickle-able
    # objects.
    #
    # When all references to this proxy object have
    # been garbage-collected, the server process is shut down.
    # Usually user doesn't need to worry about it.
    # In order to proactively shut down the server process,
    # delete (all references to) the proxy object.
    #
    # `ServerProcess.start()` can be used multiple times
    # to have multiple shared objects, which reside in diff
    # processes and are independent of each other.

    @classmethod
    def start(cls, *args, **kwargs):
        BaseManager.register(
            cls.__name__,
            cls,
        )
        manager = BaseManager()
        manager.start()  # pylint: disable=consider-using-with
        obj = getattr(manager, cls.__name__)(*args, **kwargs)
        return obj




import asyncio
import logging
import multiprocessing as mp
import queue
import time
from abc import ABCMeta, abstractmethod
from multiprocessing import synchronize
from typing import List, Type, Tuple, Sequence, Dict, Union

import psutil  # type: ignore

from ._mperror import MPError

logger = logging.getLogger(__name__)


class TimeoutError(Exception):
    pass


class EnqueueTimeout(TimeoutError):
    pass


class TotalTimeout(TimeoutError):
    pass


class Servlet(metaclass=ABCMeta):
    # Typically a subclass needs to enhance
    # `__init__` and implement `process`.

    @classmethod
    def run(cls, *,
            q_in: mp.Queue,
            q_out: mp.Queue,
            q_err: mp.Queue,
            q_in_lock: synchronize.Lock,
            cpus: Sequence[int] = None,
            **init_kwargs):
        if cpus:
            psutil.Process().cpu_affinity(cpus=cpus)
        obj = cls(**init_kwargs)
        obj.start(q_in=q_in,
                  q_out=q_out,
                  q_err=q_err,
                  q_in_lock=q_in_lock)

    def __init__(self, *,
                 batch_size: int = None,
                 batch_wait_time: float = None,
                 silent_errors: Tuple[Type[Exception], ...] = None):
        # `batch_wait_time`: seconds, may be 0.
        self.batch_size = batch_size or 0
        self.batch_wait_time = batch_wait_time or 0
        self.name = f'{self.__class__.__name__}--{mp.current_process().name}'
        self.silent_errors = silent_errors

    def _start_single(self, *, q_in, q_out, q_err):
        batch_size = self.batch_size
        silent_errors = self.silent_errors
        while True:
            uid, x = q_in.get()
            try:
                if batch_size:
                    y = self([x])[0]
                else:
                    y = self(x)
                q_out.put((uid, y))

            except Exception as e:
                if not silent_errors or not isinstance(e, silent_errors):
                    logger.info(e)
                # There are opportunities to print traceback
                # and details later using the `MPError`
                # object. Be brief on the logging here.
                err = MPError(e)
                q_err.put((uid, err))

    def _start_batch(self, *, q_in, q_out, q_err, q_in_lock):
        batch_size = self.batch_size
        batch_wait_time = self.batch_wait_time
        silent_errors = self.silent_errors
        perf_counter = time.perf_counter

        batch_size_max = -1
        batch_size_min = 1000000
        batch_size_total = 0
        n_batches = 0

        while True:
            batch = []
            uids = []
            n = 0
            with q_in_lock:
                uid, x = q_in.get()
                batch.append(x)
                uids.append(uid)
                n += 1

                wait_until = perf_counter() + batch_wait_time
                while n < batch_size:
                    time_left = wait_until - perf_counter()
                    try:
                        uid, x = q_in.get(timeout=time_left)
                    except queue.Empty:
                        break

                    batch.append(x)
                    uids.append(uid)
                    n += 1

            batch_size_max = max(batch_size_max, n)
            batch_size_min = min(batch_size_min, n)
            batch_size_total += n
            n_batches += 1
            if n_batches % 1000 == 0:
                batch_size_mean = batch_size_total / n_batches
                logger.info('batch size stats (count, max, min, mean): %d, %d, %d, %.1f',
                            n_batches, batch_size_max, batch_size_min, batch_size_mean)

            try:
                results = self(batch)
            except Exception as e:
                if not silent_errors or not isinstance(e, silent_errors):
                    logger.info(e)
                err = MPError(e)
                for uid in uids:
                    q_err.put((uid, err))
            else:
                for uid, y in zip(uids, results):
                    q_out.put((uid, y))

    def start(self, *, q_in, q_out, q_err, q_in_lock):
        q_err.put('ready')
        logger.info('%s started', self.name)
        if self.batch_size > 1:
            self._start_batch(q_in=q_in, q_out=q_out, q_err=q_err,
                              q_in_lock=q_in_lock)
        else:
            self._start_single(q_in=q_in, q_out=q_out, q_err=q_err)

    @abstractmethod
    def __call__(self, x):
        # `x`: a single element if `self.batch_size == 0`;
        # else, a list of elements.
        # When `batch_size == 0`, hence `x` is a single element,
        # return corresponding result.
        # When `batch_size > 0`, return list of results
        # corresponding to elements in `x`.
        raise NotImplementedError


class Server:
    MP_CLASS = mp
    # This class attribute is provided because in some cases
    # one may want to use `torch.multiprocessing`, which is
    # a drop-in replacement for the standard `multiprocessing`
    # with some enhancements related to data sharing between
    # processes.

    def __init__(self,
                 max_queue_size: int = None,
                 cpus: Sequence[int] = None):
        self.max_queue_size = max_queue_size or 1024
        self._q_in_out: List[mp.Queue] = [
            self.MP_CLASS.Queue(self.max_queue_size)]
        self._q_in_lock: List[synchronize.Lock] = []
        self._q_err: mp.Queue = self.MP_CLASS.Queue(self.max_queue_size)

        self._uid_to_futures: Dict[int, asyncio.Future] = {}
        self._t_gather_results: asyncio.Task = None  # type: ignore
        self._servlets: List[mp.Process] = []

        self.loop = asyncio.get_running_loop()

        if cpus:
            psutil.Process().cpu_affinity(cpus=cpus)

        self._started = False

    def add_servlet(self,
                    servlet: Type[Servlet],
                    *,
                    cpus=None,
                    workers: int = None,
                    **init_kwargs):
        # `servlet` is the class object, not instance.
        assert not self._started
        q_in = self._q_in_out[-1]
        q_in_lock = self.MP_CLASS.Lock()
        self._q_in_lock.append(q_in_lock)

        q_out: mp.Queue = self.MP_CLASS.Queue(self.max_queue_size)
        self._q_in_out.append(q_out)

        n_cpus = psutil.cpu_count(logical=True)

        if workers:
            # Number of workers is specified.
            # `cpus` specifies the cores for each worker;
            # can be `None` or `List[int]`.
            assert workers > 0
            cpus = [cpus for _ in range(workers)]
        else:
            if cpus is None:
                # Create one worker, not pinned to any core.
                cpus = [None]
            else:
                assert isinstance(cpus, list)
                # Create as many processes as the length of `cpus`.
                # Each element of `cpus` specifies cpu pinning for
                # one process. `cpus` could contain repeat numbers,
                # meaning multiple processes can be pinned to the same
                # cpu.
                # This provides the ultimate flexibility, e.g.
                #    [[0, 1, 2], [0], [2, 3], [4, 5, 6], None]

        for cpu in cpus:
            if cpu is None:
                logger.info('adding servlet %s', servlet.__name__)
            else:
                if isinstance(cpu, int):
                    cpu = [cpu]
                assert all(0 <= c < n_cpus for c in cpu)
                logger.info('adding servlet %s at CPU %s',
                            servlet.__name__, cpu)

            self._servlets.append(
                self.MP_CLASS.Process(
                    target=servlet.run,
                    name=f'servlet-{cpu}',
                    kwargs={
                        'q_in': q_in,
                        'q_out': q_out,
                        'q_err': self._q_err,
                        'cpus': cpu,
                        'q_in_lock': q_in_lock,
                        **init_kwargs,
                    },
                )
            )

    def start(self):
        assert self._servlets
        assert not self._started
        n = 0
        for m in self._servlets:
            m.start()
            n += 1
        k = 0
        while k < n:
            z = self._q_err.get()
            assert z == 'ready'
            k += 1
            print(f"servlet processes ready: {k}/{n}")

        self._t_gather_results = asyncio.create_task(self._gather_results())
        self._started = True

    def stop(self):
        if not self._started:
            return
        if self._t_gather_results is not None and not self._t_gather_results.done():
            self._t_gather_results.cancel()
            # self._t_gather_results = None
        for m in self._servlets:
            # if m.is_alive():
            m.terminate()
            m.join()
        self._started = False

        # Reset CPU affinity.
        psutil.Process().cpu_affinity(cpus=[])

    def __del__(self):
        self.stop()

    async def _gather_results(self):
        q_out = self._q_in_out[-1]
        q_err = self._q_err
        futures = self._uid_to_futures
        while True:
            while not q_out.empty():
                uid, y = q_out.get_nowait()
                fut = futures.pop(uid, None)
                if fut is None:  # timed-out in `__call__`.
                    continue
                try:
                    fut.set_result(y)
                except asyncio.InvalidStateError:
                    if fut.cancelled():
                        logger.warning('Future object is already cancelled')
                # No sleep. Get results out of the queue as quickly as possible.

            while not q_err.empty():
                uid, err = q_err.get_nowait()
                fut = futures.pop(uid, None)
                if fut is None:  # timed-out in `__call__`
                    logger.info(
                        'got error for an already-cancelled task: %r', err)
                    continue
                try:
                    fut.set_exception(err)
                except asyncio.InvalidStateError:
                    if fut.cancelled():
                        logger.warning('Future object is already cancelled')
                # No sleep. Get results out of the queue as quickly as possible.

            await asyncio.sleep(0.0013)

    async def __call__(self,
                       x,
                       *,
                       enqueue_timeout: Union[int, float] = None,
                       total_timeout: Union[int, float] = None,
                       ):
        if enqueue_timeout is None:
            enqueue_timeout = 10
        elif enqueue_timeout < 0:
            enqueue_timeout = 0
        if total_timeout is None:
            total_timeout = max(100, enqueue_timeout * 10)
        else:
            assert total_timeout > 0, "total_timeout must be > 0"
        if enqueue_timeout > total_timeout:
            enqueue_timeout = total_timeout

        loop = self.loop
        fut = loop.create_future()
        uid = id(fut)
        self._uid_to_futures[uid] = fut
        q_in = self._q_in_out[0]

        time0 = loop.time()
        time1 = time0 + enqueue_timeout
        time2 = time0 + total_timeout

        while True:
            try:
                q_in.put_nowait((uid, x))
            except queue.Full:
                timenow = loop.time()
                if timenow < time1:
                    await asyncio.sleep(0.00089)
                else:
                    fut.cancel()
                    del self._uid_to_futures[uid]
                    raise EnqueueTimeout(f'waited {timenow - time0} seconds')
            else:
                break

        try:
            await asyncio.wait_for(fut, timeout=time2 - loop.time())
        except asyncio.TimeoutError:
            # `fut` is now cancelled.
            if uid in self._uid_to_futures:
                # `uid` could have been deleted by
                # `_gather_results` during very subtle
                # timing coincidence.
                del self._uid_to_futures[uid]
            raise TotalTimeout(f'waited {loop.time() - time0} seconds')
        else:
            return fut.result()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.stop()



__version__ = '0.8.6'

from ._mpservice import Server, Servlet
from ._mpservice import TimeoutError, EnqueueTimeout, TotalTimeout
from ._mperror import MPError

__all__ = [
    'Servlet', 'Server',
    'MPError',
    'TimeoutError', 'EnqueueTimeout', 'TotalTimeout',
]
